/************************************************************************
 *        LPT-RS485 
 *            moxa102_com
 *    
*************************************************************************/
#include <linux/errno.h>
#include <linux/major.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
//#include <asm/io.h>
//#include <linux/spinlock.h>        // 2.4 ???
//#include <asm/atomic.h>

#include <linux/ioport.h>
#include <linux/parport.h>
#include <asm/irq.h>
#include <linux/wait.h>            // 2.4 ???

#include <mcua.h>
#include <lpt.h>
#include <rs485.h>
#include <rserrno.h>

#include "rs485_fifo.h"

extern    int io1, io2;
extern    int irq1, irq2;

static volatile int data_ready = 0;
DECLARE_WAIT_QUEUE_HEAD(tr_queue);        // 2.4 ???

int rs485_getIO(int port) {
    if(port==2) return _IO_LPT2;
    return _IO_LPT1;
}

int rs485_getIRQ(int port) {
    if(port==2) return _IRQ_LPT2;
    return _IRQ_LPT1;
}

#ifndef KERN_2_6
static int rs_irq;
#endif
int rs485_get_port(int irq) {
#ifndef KERN_2_6
    rs_irq = irq;
#endif
    if(irq == irq1) return 1;
    else return 2;
}

void rs485_allow_interrupt_card(RS485INFO *pinfo) {
//   
    u8 a = inb(pinfo->base +EPP_CONTROL);
#ifdef __M_BARRIER__
    rmb();
#endif
    outb( a | EPP_INT_ENABLE, pinfo->base +EPP_CONTROL );
#ifdef __M_BARRIER__
    wmb();
#endif
}

void rs485_stop_interrupt_card(RS485INFO *pinfo) {
//   
    u8 a = inb(pinfo->base +EPP_CONTROL);
#ifdef __M_BARRIER__
    rmb();
#endif
    outb( a & ~EPP_INT_ENABLE, pinfo->base +EPP_CONTROL );
#ifdef __M_BARRIER__
    wmb();
#endif
}

inline void rs485_read_enable(RS485INFO *pinfo) {
    u8 a = inb(pinfo->base +EPP_CONTROL);
#ifdef __M_BARRIER__
    rmb();
#endif
    outb( a | ENABLE_BI_DIRECT,
                pinfo->base +EPP_CONTROL );
#ifdef __M_BARRIER__
    wmb();
#endif
}

inline void rs485_transmit_enable(RS485INFO *pinfo) {
    u8 a = inb(pinfo->base +EPP_CONTROL);
#ifdef __M_BARRIER__
    rmb();
#endif
    outb( a & ~ENABLE_BI_DIRECT,
                        pinfo->base +EPP_CONTROL );
#ifdef __M_BARRIER__
    wmb();
#endif
}

#define Fosc        14746
#define Minbaudrate    115
#define Maxbaudrate 1840

int ubrr_to_rate(int baudrate_code) {
    int a;
    a = baudrate_code + 1;
    if(a == 0) return 0;
    return( baudrate_code == 0x80 ? Fosc/8 : Fosc/16/a );
}

int rate_to_ubrr(int baudrate) {
    int baudrate_code;
    baudrate = baudrate < Minbaudrate ? Minbaudrate : baudrate;
    baudrate = baudrate > Maxbaudrate ? Maxbaudrate : baudrate;

    baudrate_code = baudrate > 1200 ? 0x80 : Fosc/16/baudrate-1;
//    baudrate_code = baudrate > 1200 ? 0x80 : iround(Fosc/16.0/baudrate)-1;
//VK       1800   -1 ( 0xFF ) 
//    
    return baudrate_code;
}

int rs485_set_baudrate(RS485INFO *pinfo) {
    int baudrate_code;
    baudrate_code = rate_to_ubrr(pinfo->baudrate);
    pinfo->baudrate = ubrr_to_rate(baudrate_code);
    PDEBUG("baudrate=%d\n",pinfo->baudrate);

    if( pinfo->interface_type == CONV_VERS1 ) {
//         udelay(10000);   
        pinfo->outbuffer.size = 9;
        pinfo->outbuffer.cache[pinfo->outbuffer.size] = 0;
        pinfo->outbuffer.cache[0] = 0x1f;
        pinfo->outbuffer.cache[1] = 0x07;
        pinfo->outbuffer.cache[2] = 0x1f;
        pinfo->outbuffer.cache[3] = 0x00;
        pinfo->outbuffer.cache[4] = baudrate_code;
        pinfo->outbuffer.cache[5] = 0x73;
        pinfo->outbuffer.cache[6] = 0x00;
        pinfo->outbuffer.cache[7] = 0x01;
        pinfo->outbuffer.cache[8] = 0x6c;
        start_transmitter(pinfo);
    }
    if( pinfo->interface_type == CONV_VERS2 ) {
        pinfo->outbuffer.size = 3;
        pinfo->outbuffer.cache[pinfo->outbuffer.size] = 0;
        pinfo->outbuffer.cache[0] = 0x1f;
        pinfo->outbuffer.cache[1] = 0x43;
        pinfo->outbuffer.cache[2] = baudrate_code;
          start_transmitter(pinfo);
    }
//    udelay(10000);
    return SUCCESS;
}

int rs485_reset(RS485INFO *pinfo) {
    u8 i;

    int ret = SUCCESS;
    ret = rs485_test_device(pinfo); // a to set_baudrate povesip komp!
    if(ret != SUCCESS) return ret;

    outb( EPP_TIMEOUT, pinfo->base +EPP_STATUS );        //  ?
#ifdef __M_BARRIER__
    wmb();
#endif
    outb( 0, pinfo->base +EPP_CONTROL );                //   !!!
#ifdef __M_BARRIER__
    wmb();
#endif
    udelay(EPP_WAIT);    // 250 mksec
    outb( EPP_RESET, pinfo->base +EPP_CONTROL );
#ifdef __M_BARRIER__
    wmb();
#endif
    udelay(EPP_WAIT);    // 250 mksec

    outb( 0x0e, pinfo->base +EPP_DATA );
#ifdef __M_BARRIER__
    wmb();
#endif
    outb( 0x1f, pinfo->base +EPP_ADDR );
#ifdef __M_BARRIER__
    wmb();
#endif
    udelay(RESP_WAIT);    // 100 mksec
// New code:
    i = inb(pinfo->base+EPP_DATA);
#ifdef __M_BARRIER__
    rmb();
#endif
      if( i == SignalACN ) {
          pinfo->interface_type = CONV_VERS2;
          outb( 0x87, pinfo->base +EPP_DATA ); //For new
#ifdef __M_BARRIER__
        wmb();
#endif
        outb( 0x1f, pinfo->base +EPP_ADDR );
#ifdef __M_BARRIER__
        wmb();
#endif
        udelay(RESP_WAIT);    // 100 mksec
        inb(pinfo->base+EPP_DATA);
#ifdef __M_BARRIER__
        rmb();
#endif
        PDEBUG("Converter code vers.2\n");
    } else  if (i == SignalACY) {
        pinfo->interface_type = CONV_VERS1;
        PDEBUG("Converter code vers.1\n");
    } else {
        PDEBUG( "reset: NO POWER\n" );
        return -ERSSHUTDOWN;    // na nekotorykh kompakh test_device ne srabatyvayet
                        
    }

    udelay(EPP_WAIT);

    pinfo->status = ACTIVE_DATA;
    rs485_read_enable(pinfo);
// Tut visnet komp naskvoz', esli ne vkliuchen pribor ili ne vkliuchen pribor ili net konvertera
// (Ne liuboj komp., no mnogie)

    rs485_allow_interrupt_card(pinfo);

    rs485_set_baudrate(pinfo);
    return ret;
}

int rs485_test_device(RS485INFO *pinfo) {

    u8 iir = inb(pinfo->base+EPP_STATUS);
#ifdef __M_BARRIER__
    rmb();
#endif
    if( iir & PARPORT_STATUS_PAPEROUT ) {
        PDEBUG( "NO Converter!\n");
        return -ERSNOINTERFACE;
    }

    if( iir & PARPORT_STATUS_SELECT ) {
        PDEBUG( "NO POWER\n" );
        return -ERSSHUTDOWN;
    }

    if( iir & EPP_TIMEOUT ) {
        PDEBUG( "EPP TIMEOUT\n" );
        outb( EPP_TIMEOUT, pinfo->base +EPP_STATUS );
#ifdef __M_BARRIER__
        wmb();
#endif
        return -ERSISATMOUT;
    }
    return SUCCESS;
}

int start_transmitter(RS485INFO *info) {
/* 
 *   
 *   < 0
 *   ,   info->.. (    c)...
*/
#ifndef KERN_2_6
    int ready=0;
    int resp_tmout=0;
#endif
    long remain_tm;
    const long timeout = HZ * RESP_TIMEOUT / 1000L;    // JIFFIES

#ifdef __DEBUG_SIGNAL__
    outb(TXENABLE, 0x3f8 +UART_MCR);    // to COM1
#endif

    data_ready = 0;    // not ready
    rs485_transmit_enable(info);

    info->usCountByteOut = 0;    //    0
//      ,     ,
//         ,   ...
    while( 1 ) {

        info->usCountByteOut ++;
        if( info->usCountByteOut >= info->outbuffer.size ) {
            break;
        }

        //    ( )
        outb(info->outbuffer.cache[info->usCountByteOut], // 1-, ,
                info->base +EPP_DATA);                        //  
#ifdef __M_BARRIER__
        wmb();
#endif
    }
    //   (     )
    outb( info->outbuffer.cache[0], info->base +EPP_ADDR );
#ifdef __M_BARRIER__
    wmb();
#endif
    info->status = PASSIVE_DATA;

    PDEBUGG( "send: %d bytes, QUIT\n",info->usCountByteOut );

#ifndef KERN_2_6
    rs485_read_enable(info);

//    ,    

    rs485_set_starttime();
    resp_tmout = 0;
    while( 1 ) {
//        rs485_stop_interrupts(); 
        disable_irq(rs_irq); // protect race condition with handler
        //   mem  ,       
//  mutex    ???
        ready = data_ready;
        enable_irq(rs_irq);
//        rs485_allow_interrupts();

        if(ready) break;
        if( rs485_timeout(RESP_TIMEOUT) ) {
            resp_tmout = 1;
            break;
        }
        udelay(RESP_WAIT);
    }
#ifdef __DEBUG_SIGNAL__
    outb(RXENABLE, 0x3f8 +UART_MCR);    // to COM1
#endif
    if(resp_tmout) {
        PDEBUG( "transmit: timeout!\n" );
        return -ERSNOCONVREPLY;
    } else {
        return(info->passive_inbuf.size);
    }
#else
    remain_tm = wait_event_interruptible_timeout(tr_queue, data_ready != 0, timeout);
//       read_enable  ???  wait_event_interruptible_timeout ?
    rs485_read_enable(info);
#ifdef __DEBUG_SIGNAL__
    outb(RXENABLE, 0x3f8 +UART_MCR);    // to COM1
#endif
    if(signal_pending(current)) {
        return -ERESTARTSYS;
    }
    if( remain_tm == 0 ) {
        PDEBUG( "transmit: timeout!\n" );
        return -ERSNOCONVREPLY;
    }
    return(info->passive_inbuf.size);    //    
#endif
}


void receiv_data(RS485INFO * info) {

/*
 * -----------------------------------------------------------
 *           .
 *          info->status == PASSIVE_DATA,    
 *         .   : [length][data...]  [signal]
 *          : [address][length][data...]
 * -----------------------------------------------------------
*/

    u8 iir = 0;
    int count_byte = 0;     //        (  )
    struct _STACK_BUFFER_INP_ *current_inbuf;

#ifdef __DEBUG_TIMING__
    rs485_set_starttime();
#endif

#ifdef __DEBUG_SIGNAL__
    outb(TXENABLE, 0x3f8 +UART_MCR);    // to COM1
#endif

    // ,   -  
START:    
    iir = inb(info->base +EPP_STATUS);
#ifdef __M_BARRIER__
    rmb();
#endif
    if( (iir & PARPORT_STATUS_ACK) == 0 ) {
        goto QUIT_HANDLER;
    }

    info->signal = 0;
    info->usCountByteInp = 0; //   
    if( info->status == PASSIVE_DATA ) {
        current_inbuf = &(info->passive_inbuf);
        current_inbuf->size = 0;    // yesli ne smozhem nichego polozhit'
        info->lSection = SECT_LENGTHBYTE;
    } else {
        current_inbuf = &(info->active_inbuf);
        info->lSection = SECT_ADDRESSBYTE;
    }

    while( 1 ) {

        info->tmp_byte = inb( info->base +EPP_DATA );

        count_byte++;
        if(count_byte > CACHE_MAXSIZE) {
                PDEBUG("long handler\n");
                goto QUIT_HANDLER;
        }
                                                                                
        //  -  "-"
/*
        iir = inb(info->base +EPP_STATUS);
        mb();
        if( iir & EPP_TIMEOUT ) {

            PDEBUG( "receiver: timeout\n" );
            // reset:
            outb(EPP_TIMEOUT, info->base +EPP_STATUS);
            // wmb();
            if( (iir & PARPORT_STATUS_ACK) == 0 ) {
                PDEBUG( "receiver: timeout: No data in converter\n" );
                goto QUIT_HANDLER;    // no data in converter
            }
            info->tmp_byte = inb( info->base +EPP_DATA );
            mb();
            iir = inb(info->base +EPP_STATUS);
            mb();
            if( iir & EPP_TIMEOUT ) {
                PDEBUG( "receiver: repeat. timeout, quit\n" );
                goto QUIT_HANDLER;
            }
        }
*/

        switch(info->lSection) {

        case SECT_ADDRESSBYTE:

            //    addressbyte  cc !

            current_inbuf->cache[info->usCountByteInp] = info->tmp_byte;
            info->usCountByteInp ++;
            current_inbuf->adr  = info->tmp_byte;

            //      
            info->lSection = SECT_LENGTHBYTE;

            break;

        case SECT_LENGTHBYTE:

            //      ?
            if(info->tmp_byte >= 0x20) { //  
                if(info->tmp_byte < CACHE_MAXSIZE) { // length error
                    PDEBUG( "receiver: length error! lenght = %x\n", info->tmp_byte);
                } else {    // signal
                    if(info->status == PASSIVE_DATA ) {
                        current_inbuf->size = info->tmp_byte;
//  info->signal  ,      ,    
                        data_ready = 1;
                        wake_up(&tr_queue);
                        info->status = ACTIVE_DATA;
                    } else {    // Active data:
                        PDEBUG( "active receiver: signal = %x received!\n", info->tmp_byte);
                        info->signal = (char)info->tmp_byte;
                        rs485_put_data(info);
                    }
                }
                goto START;
            }

            current_inbuf->cache[info->usCountByteInp] = info->tmp_byte;
            info->usCountByteInp ++;

            if( info->status == PASSIVE_DATA ) {
                current_inbuf->size = info->tmp_byte+1;
            } else {
                // +   
                current_inbuf->size = info->tmp_byte+2;
            }

            info->lSection = SECT_DATABYTES;

            break;

        case SECT_DATABYTES:            

            //  

            current_inbuf->cache[info->usCountByteInp] = info->tmp_byte;
            info->usCountByteInp ++;
            //       -   
            if( info->usCountByteInp < current_inbuf->size ) break;

            //  

            if( info->status == ACTIVE_DATA ) {
//                    printk("size=%d inp=%d\n",current_inbuf->size,info->usCountByteInp);
                //    FIFO
#ifdef __DEBUG_TIMING__
                if( rs485_timeout(100L) ) {    // 1 millisecond = 1000 usec
                    RS_PRINT( "Long packet: > 100 usec!\n" );
                }
#endif

                  rs485_put_data(info);
//          c... 
                goto START;    //    - ...
            } else {

                //      

                data_ready = 1;
                wake_up(&tr_queue);    // 2.4 ???
                info->status = ACTIVE_DATA;
                goto START;
            }

        }
    } //   

QUIT_HANDLER:

#ifdef __DEBUG_SIGNAL__
    outb(RXENABLE, 0x3f8 +UART_MCR); // to COM1
#endif

    return;
}
