/*-------------------------------------------------
 *	Kernel version 2.4.20 - 2.6.8, 2.6.13
 *	RS485 (Moxa132, Moxa102, COM-port, LPT-port :
 * 
 *	Andre Zaitsev, Olga Voziakova
 *	ovoz@sai.msu.ru
 *	$History:
 *
 *---------------------------------------- */
#include <linux/module.h>
#include <linux/types.h>
#include <linux/kdev_t.h>
#include <linux/fs.h>

#include <linux/errno.h>
#include <linux/slab.h>
#include <asm/uaccess.h>	//  copy_from_user
#include <linux/ioport.h>
#include <linux/delay.h>	// udelay

#include <mcua.h>
#ifdef KERN_2_6
#include <linux/cdev.h>
#endif
#include <rserrno.h>
#ifdef __CP132__
	#include <moxa132.h>
#endif
#ifdef __CP102__
	#include <moxa102.h>
        #include "moxa102_com.h"
#endif
#ifdef __COM__
	#include <com.h>
        #include "moxa102_com.h"
#endif
#ifdef __LPT__
	#include <lpt.h>
#endif

#include <rs485.h>
#include "rs485_tools.h"
#include "rs485_fifo.h"
#include "rs485_irq.h"

#ifdef __RS485_PROC__
extern int rs485_create_proc(void);
extern void rs485_remove_proc(void);
#endif

static unsigned int major;
#ifdef KERN_2_6
static dev_t dev;
static struct cdev *cdev;
static unsigned int rs_nr_dev = 2;
static unsigned int minor = 0;
#endif

// ,     insmod
int io1 = 0;			//  1  -
int io2 = 0;			//  2  -
// static 2.4 ?
int irq1 = 0;			//   port1
// static 2.4
int irq2 = 0;			//   port2
							// 2-COM2 ( PORT2  MOXA102)...
static int baudrate1 = 0;		//   port1
static int baudrate2 = 0;		//   port2

#ifndef __PCI__	//   ISA ( Moxa C102  COM-)
#else
    extern  int _find_PCImoxa(void);
#endif

struct file_operations	rs485_fops;
// static 2.4 ?
RS485INFO		*info_port1, *info_port2;

void rs485_release_resource(void) {
	if( io2 != 0 ) release_region(io2, ADDRESS_REGION_SIZE);
	if( io1 != 0) release_region(io1, ADDRESS_REGION_SIZE);
	if( irq2 != 0 ) rs485_clean_irq((unsigned int) irq2);
	if( (irq1 != 0) && (irq1 != irq2) ) rs485_clean_irq((unsigned int) irq1);
	return;
}

int rs485_init(void) {		//  addresov i preryvanij
	int err = 0;
//	int major = RS485_MAJOR;
	
#ifdef __PCI__
	int res = _find_PCImoxa();
	if(res != SUCCESS) {
		RS_PRINT(KERN_INFO "%s: NO CARD!\n", ADAPTER_ID);
		return res;
	}
#endif

#ifndef KERN_2_6

	if( (err=check_region(io1, ADDRESS_REGION_SIZE)) != 0 ) {
		RS_PRINT( KERN_CRIT "%s: I/O port 0x%08lX in use.\n",
					ADAPTER_ID, (unsigned long) io1);
		io1 = 0;
		irq1 = 0;
			printk("io1=%x failed\n",io1);
	} else {
		request_region(io1, ADDRESS_REGION_SIZE, ADAPTER_ID);
	}
	
	if( (err=check_region(io2, ADDRESS_REGION_SIZE)) != 0 ) {
		RS_PRINT( KERN_CRIT "%s: I/O port 0x%08lX in use.\n",
					ADAPTER_ID, (unsigned long) io2);
		io2 = 0;
		irq2 = 0;
			printk("io2=%x failed\n",io2);
	} else {
		request_region( io2, ADDRESS_REGION_SIZE, ADAPTER_ID );
	}	
#else
	if(!request_region(io1, ADDRESS_REGION_SIZE, ADAPTER_ID)) {
		RS_PRINT( KERN_CRIT "%s: I/O port 0x%08lX in use.\n",
					ADAPTER_ID, (unsigned long) io1);
		io1 = 0;
		irq1 = 0;
		err = -EACCES;
	}
	if(!request_region(io2, ADDRESS_REGION_SIZE, ADAPTER_ID)) {
		RS_PRINT( KERN_CRIT "%s: I/O port 0x%08lX in use.\n",
					ADAPTER_ID, (unsigned long) io2);
		io2 = 0;
		irq2 = 0;
		err = -EACCES;
	}	
#endif
	if( (io1 == 0) && (io2 == 0) ) return(err);

	//  

	if(irq1 != 0) if( (err=rs485_request_irq(irq1)) != 0 ) {
		if( io1 != 0 ) release_region(io1, ADDRESS_REGION_SIZE);
		io1 = 0;
		irq1 = 0;
	}
	if((irq2 != 0) && (irq1 != irq2)) {
		if( (err=rs485_request_irq(irq2)) != 0 ) {
			if( io2 != 0 ) release_region(io2, ADDRESS_REGION_SIZE);
			io2 = 0;
			irq2 = 0;
		}
	}

	if( (irq1 == 0) && (irq2 == 0) ) return(err);

#ifndef KERN_2_6
	if( (major=register_chrdev(0, ADAPTER_ID, &rs485_fops)) <= 0) {
		RS_PRINT( KERN_CRIT "%s: unable to get major\n", ADAPTER_ID);
		rs485_release_resource();
		return(major);	//err=major
	} else RS_PRINT(KERN_INFO "major=%d\n",major);
#else
	if( (err=alloc_chrdev_region(&dev, minor, rs_nr_dev, ADAPTER_ID)) < 0 ) {
		RS_PRINT( KERN_CRIT "%s: unable to get major\n", ADAPTER_ID);
		rs485_release_resource();
		return err;
	}
	major = MAJOR(dev);
	RS_PRINT(KERN_INFO "major=%d\n",major);

	cdev = cdev_alloc();
	cdev->ops = &rs485_fops;
	cdev->owner = THIS_MODULE;
	err = cdev_add(cdev, dev, rs_nr_dev);
	if (err) {
		RS_PRINT(KERN_WARNING "rs485: device can't added to the system\n");
		return err;
	}
#endif
	return SUCCESS;
}

void rs485_blnumber_reset(RS485INFO *info) {
//	  
	int i;
	for( i=0; i<MAX_TASK_DEVICE; i++ ) {	
		info->device[i].number = 0;
		info->device[i].round_count_prev = -1;
	}
	info->passive_inbuf.adr = 0;
	info->active_inbuf.adr = 0;
}

/* ------------------------------------------------------------
 *    
 * -----------------------------------------------------------*/

int rs485_open(struct inode *inode, struct file *pfile) {

//    .     

	RS485INFO *info;
	int ret=0;
	int imaj, imin;
	
#ifndef KERN_2_6
	switch(MINOR(inode->i_rdev)) {		//   
#else
	switch(iminor(inode)) {
#endif
		case 0:
			if( io1 == 0 ) return(-ERSBUSY);
			info = info_port1;			//   1
			break;

		case 1:
			if( io2 == 0) return(-ERSBUSY);
			info = info_port2;			//   2
			break;

		default: return(-ERSNOPORT);
	}

	pfile->private_data = info;

#ifndef KERN_2_6
			imaj = MAJOR(inode->i_rdev);
			imin = MINOR(inode->i_rdev);
#else
			imaj = imajor(inode);
			imin = iminor(inode);
#endif
	PDEBUG("%s: open device %d, %d\n", ADAPTER_ID, imaj, imin);

	if( info->use_mcua == 0 ) {	//   

//		if( pfile->f_flags & O_NONBLOCK ) info->fifo.nonblock = 1;
//		else info->fifo.nonblock = 0;
		
		rs485_fifo_init(info);		//  FIFO   
// test_device sidit v reset(), t.k. v com nado snachala delat'
// reset, a potom test_device, a v lpt naoborot.		
		ret = rs485_reset(info); 
		if( ret < 0 ) return(ret);

	} else if( info->use_mcua >= MAX_USE_STREAM ) {
// 		    
		return(-ERSBUSY);
	}

	info->use_mcua ++;
#ifndef KERN_2_6
	MOD_INC_USE_COUNT;
#endif
	return SUCCESS;
}

int rs485_close(struct inode *inode, struct file *pfile) {
/*     
 *      
 */

	RS485INFO *info = pfile->private_data;

	if( down_interruptible(&info->sem)) {		// 2.4 ???
		return -ERESTARTSYS;
	}
#ifndef KERN_2_6
	MOD_DEC_USE_COUNT;
#endif
	info->use_mcua --;

	if( info->use_mcua == 0 ) {
		//      
		rs485_stop_interrupt_card(info);
#ifndef __LPT__
		rs485_clean_reg(info);
#endif
    }
//#ifndef KERN_2_6
//	down_flags(FMCUA_OPEN, &info->lFlags);
    udelay(SEM_WAIT);
    up(&info->sem);				// 2.4 ???
    return SUCCESS;
}

ssize_t rs485_write( struct file *fp, const char *buf, size_t size, loff_t *offset ) {

/*      ( NE   write )
 *    < 0  	 
 *				> CACHE_MAXSIZE	  
 *				< CACHE_MAXSIZE	   ,
 * 					zapisyvaemykh v pol'zovatel'skij buf
 *				CRC ne peredayotsia pol'zovateliu.
 */	

	RS485INFO	*info = fp->private_data;
#ifndef __LPT__
	unsigned char crc_sum;
#endif	
	int ret = 0;

	if( size > CACHE_MAXSIZE ) {
		PDEBUG( "write: size=%d is too match!\n", size);
		return -ERSFORMAT;
	}
	if( down_interruptible(&info->sem)) {		// 2.4 ???
		return -ERESTARTSYS;
	}

#ifdef __COM__
	if( (ret=rs485_test_interface(info)) < 0 ) goto OUT;
#else
	if( (ret=rs485_test_device(info)) < 0 ) goto OUT;
#endif
	//   
	copy_from_user(info->outbuffer.cache, buf, size);
	info->outbuffer.cache[size] = 0;
	info->outbuffer.size = size;

#ifndef __LPT__
	//   
	// LPT-RS485 converter   CRC

	crc_sum = crc(info->outbuffer.cache, info->outbuffer.size);
	info->outbuffer.cache[size] = crc_sum;
#endif
		//  
	ret = start_transmitter(info);
	if( (ret > 0) && (ret <= CACHE_MAXSIZE) ) {
		copy_to_user((char*)buf, info->passive_inbuf.cache, ret);
	}
OUT:
        udelay(SEM_WAIT);
	up(&info->sem);
	return ret;
}

ssize_t rs485_read( struct file *fp, char *buf,
			size_t count, loff_t *offset ) {
/* 
 * Read data fifo
 */	

	char count0;
	RS485INFO	*info = fp->private_data;
	int ret;

	if( down_interruptible(&info->sem)) {		// 2.4 ???
		return -ERESTARTSYS;
	}

	ret = rs485_data_fifo_read( info, buf, 1 );
	if( ret > 0 ) {
		get_user(count0,buf);	// buf in User space!!!
		if( count0 < (count-1) ) count = count0;
		else count--;
		ret = rs485_data_fifo_read( info, buf+1, count );
	}
        udelay(SEM_WAIT);
	up(&info->sem);
	return ret;
}

int rs485_ioctl(struct inode *inode, struct file *fp,
			unsigned int cmd, unsigned long arg) {

	int err = SUCCESS;
	RS485INFO *rs485;
	
	if(_IOC_TYPE(cmd) != RS_IOC_MAGIC) {		// 2.4 ???
		PDEBUG("ioctl_error!\n");
		return -ENOTTY;
	}
	if(_IOC_NR(cmd) > RS_IOC_MAXNR) {
		PDEBUG("ioctl_error!\n");
		return -ENOTTY;
	}
#ifdef KERN_2_6
	if(_IOC_DIR(cmd) & _IOC_READ) {
		err = !access_ok(VERIFY_WRITE, (void __user *)arg, _IOC_SIZE(cmd));
	} else {
		err = !access_ok(VERIFY_READ, (void __user *)arg, _IOC_SIZE(cmd));
	}
#endif
	if(err != SUCCESS) {
		PDEBUG("ioctl_error!\n");
		return -EFAULT;
	}
	rs485 = fp->private_data;

	switch(cmd) {
		case RS_RESET:
			if( down_interruptible(&rs485->sem)) {
				return -ERESTARTSYS;
			}
			err = rs485_reset(rs485);
			rs485_fifo_flush(rs485);
                        udelay(SEM_WAIT);
			up(&rs485->sem);
			return err;

		case RS_FIFO_FLUSH:
			if( down_interruptible(&rs485->sem)) {
				return -ERESTARTSYS;
			}
			rs485_fifo_flush(rs485);
                        udelay(SEM_WAIT);
			up(&rs485->sem);
			return SUCCESS;

		case RS_SET_BAUDRATE:
			if( down_interruptible(&rs485->sem)) {
				return -ERESTARTSYS;
			}
                        err = rs485_test_device(rs485);
                        if(err >= 0 ) {
   				rs485->baudrate = (int)arg;
				err = rs485_set_baudrate(rs485);
                        }
                        udelay(SEM_WAIT);
			up(&rs485->sem);
			return err;

		case RS_GET_BAUDRATE:
			return rs485->baudrate;

		case RS_BLNUMBER_RESET:
			rs485_blnumber_reset(rs485);
			return SUCCESS;

		case RS_GET_VERS:
			copy_to_user((char*)arg, RS485_UTS_RELEASE, sizeof(RS485_UTS_RELEASE));
			return RS485_VERSION_CODE;

		case RS_GET_DATAFIFO_SIZE:
			return SIZE_STREAM_FIFO;

		default: break;
	}
	return -ERSNOIOCTL;
}


// ---------------------------------------------------------------------

// --------    --------------------------------
struct file_operations rs485_fops = {

	.owner		= THIS_MODULE,		// 2.4 ???
	.open		= rs485_open,
	.release	= rs485_close,
	.read		= rs485_read,
	.write		= rs485_write,
	.ioctl		= rs485_ioctl,
/*
	open:		rs485_open,		// OPEN
	release:	rs485_close,	// CLOSE 
	read:		rs485_read,		// READ
	write:		rs485_write,	// WRITE
	ioctl:		rs485_ioctl,	// IOCTL
//	flush:		NULL			// FLUSH
*/
};

/* ---------------------------------------------------------------
 *	   
 * --------------------------------------------------------------*/

void rs485_free_mem1(void) {
	kfree(info_port1);
}

void rs485_free_mem2(void) {
	kfree(info_port2);
}

void rs485_delete_card(void) {
	rs485_release_resource();
#ifndef KERN_2_6
	if ( (unregister_chrdev(major, ADAPTER_ID)) !=0){
		RS_PRINT(KERN_CRIT "%s: Error unload module\n", ADAPTER_ID);
	}
#else
	cdev_del(cdev);
	unregister_chrdev_region(dev,rs_nr_dev);
#endif
	return;
}

/* -------------------------------------------------
 *      
 *  :
 * io  <base address>
 * irq <interrupt>
 * --------------------------------------------------- */
 
MODULE_AUTHOR("Andre Zaitsev, Olga Voziakova (ovoz@sai.msu.ru)");
MODULE_DESCRIPTION("rs485 module");
#ifdef MODULE_LICENSE
	MODULE_LICENSE("GPL");
#endif

#ifdef __CP102__
    MODULE_PARM_DESC(io1, "Base port1 I/O Moxa C102 (hex), default io1=_IO_CARD1");
    MODULE_PARM_DESC(io2, "Base port2 I/O Moxa C102 (hex), default io2=_IO_CARD2");
    MODULE_PARM_DESC(irq1, "Interrupt Moxa C102, default irq1=_IRQ_CARD");
#endif
#ifdef __COM__
    MODULE_PARM_DESC(io1, "Base COM1-port (hex), default io1=_IO_COM1");
    MODULE_PARM_DESC(io2, "Base COM2-port (hex), default io2=_IO_COM2");
    MODULE_PARM_DESC(irq1, "Interrupt COM1-port, default irq1=_IRQ_COM1");
    MODULE_PARM_DESC(irq2, "Interrupt COM2-port, default irq2=_IRQ_COM2");
#endif
#ifdef __LPT__
    MODULE_PARM_DESC(io1, "Base LPT1-port (hex), default io1=_IO_LPT1");
    MODULE_PARM_DESC(io2, "Base LPT2-port (hex), default io2=_IO_LPT2");
    MODULE_PARM_DESC(irq1, "Interrupt LPT1-port,default irq1=_IRQ_LPT1");
    MODULE_PARM_DESC(irq2, "Interrupt LPT2-port,default irq2=_IRQ_LPT2");
#endif

MODULE_PARM_DESC(io1, "i");
MODULE_PARM_DESC(io2, "i");
MODULE_PARM_DESC(irq1, "i");
MODULE_PARM_DESC (irq2, "i");

//MODULE_PARM_DESC(portnumb, "PORT NUMBER, default portnumb=_PORT_NUMB");
//MODULE_PARM(portnumb, "i");

MODULE_PARM_DESC(baudrate1, "BAUD RATE UART port1, default baudrate=_BAUDRATE_1");
MODULE_PARM_DESC(baudrate2, "BAUD RATE UART port2, default baudrate=_BAUDRATE_2");
MODULE_PARM_DESC(baudrate1, "i");
MODULE_PARM_DESC(baudrate2, "i");

#ifndef KERN_2_6
EXPORT_NO_SYMBOLS;
#endif

int init_module(void) {
	int err = 0;
	int err1 = 0;
	int err2 = 0;
	char version[]=RS485_UTS_RELEASE;

//	portnumb = portnumb ? portnumb : _PORT_NUMB;	//     insmod...
	baudrate1 = baudrate1 ? baudrate1 : _BAUDRATE_1;	//	 _BAUDRATE_1,   insmod
														//	  
	baudrate2 = baudrate2 ? baudrate2 : _BAUDRATE_2;

//	PDEBUG( "PORT %d\n", portnumb );
	PDEBUG("baudrate1=%d baudrate2=%d\n", baudrate1, baudrate2);

#ifndef __PCI__
	io1 = io1 ? io1 : rs485_getIO(1);		//     insmod -
	io2 = io2 ? io2 : rs485_getIO(2);
	irq1 = irq1 ? irq1 : rs485_getIRQ(1);	//   ... 
	irq2 = irq2 ? irq2 : rs485_getIRQ(2);
#endif


//	 

	info_port1 = (RS485INFO *) kmalloc (sizeof (RS485INFO), GFP_ATOMIC);
	if(info_port1 == NULL){
		err = -ENOMEM;
		goto FAIL_1;
	}
	
	info_port2 = (RS485INFO *) kmalloc (sizeof (RS485INFO), GFP_ATOMIC);
	if(info_port2 == NULL){
		err = -ENOMEM;
		goto FAIL_2;
	}
	
	memset((void *)info_port1, 0 , sizeof(RS485INFO)); //  
	memset((void *)info_port2, 0 , sizeof(RS485INFO)); //  

	rs485_blnumber_reset(info_port1);	//   
	rs485_blnumber_reset(info_port2);	//   

	if( (err=rs485_init()) != 0 ) goto FAIL_3;

	info_port1->base = io1;
	info_port2->base = io2;

	info_port1->baudrate = baudrate1;
	info_port2->baudrate = baudrate2;

	rs485_stop_interrupt_card(info_port1);
	rs485_stop_interrupt_card(info_port2);

	info_port1->usHopInp = 0;
	info_port2->usHopInp = 0;

	init_MUTEX(&info_port1->sem);		// 2.4 ???
	init_MUTEX(&info_port2->sem);

//    

	if( ((err1=rs485_fifo_create(info_port1)) != 0)
		|| ((err2=rs485_fifo_create(info_port2)) != 0) ) {
		err = err1 ? err1 : err2;
		goto FAIL_5;
	}

	RS_PRINT(KERN_INFO "%s module, version %s\n",
				ADAPTER_ID, version);
	RS_PRINT(KERN_INFO "%s: install card on the base=0x%X, 0x%X and irq=%d, %d\n",
				ADAPTER_ID, io1, io2, irq1, irq2);

#ifdef __RS485_PROC__
	if( (err=rs485_create_proc()) ) goto FAIL_6;
#endif

	return SUCCESS;

#ifdef __RS485_PROC__
FAIL_6:
#endif
	rs485_fifo_destroy();
FAIL_5:	//	create_fifo failed
	rs485_delete_card();
FAIL_3:
	rs485_free_mem2();
FAIL_2:
	rs485_free_mem1();
FAIL_1:
	return(err);
}

void cleanup_module(void){
	rs485_fifo_destroy();
	rs485_free_mem2();
	rs485_free_mem1();
	rs485_delete_card();
#ifdef __RS485_PROC__
	rs485_remove_proc();
#endif
	return;
}
