/************************************************************************
 *        MOXA CP102
 *            cp132 1.2.3
 *    
*************************************************************************/
#include <linux/errno.h>
#include <linux/major.h>
#include <linux/delay.h>    // udelay()

#include <linux/interrupt.h>


#include <mcua.h>
#ifdef __COM__
  #include <com.h>
#else
  #include <moxa102.h>
#endif

#include <linux/ioport.h>

#include <rs485.h>
#include <rserrno.h>
#include "rs485_tools.h"
#include "rs485_fifo.h"
#include "moxa102_com.h"
#include "rs485_irq.h"
extern int irq1, irq2;
extern RS485INFO *info_port1, *info_port2;

int rs485_getIO(int port) {
#ifdef __COM__
    switch(port) {
        case 1    : return _IO_COM1;
        case 2    : return _IO_COM2;
//        case 3    : return _IO_COM3;
//        case 4    : return _IO_COM4;
        default    : return _IO_COM1;
    }
#else
    switch(port) {
        case 1    : return _IO_CARD1;
        case 2    : return _IO_CARD2;
        default    : return _IO_CARD1;
    }
#endif
}

int rs485_getIRQ(int port) {
#ifdef __COM__
    switch(port) {
        case 1    : return _IRQ_COM1;
        case 2    : return _IRQ_COM2;
//        case 3    : return _IRQ_COM3;
//        case 4    : return _IRQ_COM4;
        default    : return _IRQ_COM1;
    }
#else
    return _IRQ_CARD;
#endif
}

#ifndef __COM__
static unsigned short iir_stat=0;
// pri sleduyuchem schityvanii izmenitsia PE-bit
// (po opisaniyu moxy) => nado zapomnit' dlia receiver-a
#endif

int rs485_get_port(int irq) {
#ifdef __COM__
    if(irq == irq1) return 1;
    else return 2;
#else
//    unsigned short iir=0;
    iir_stat = inb(info_port1->base +UART_LSR);
    if( (iir_stat & UART_LSR_DR) != 0 ) return 1; //info_port2
    iir_stat = inb(info_port2->base +UART_LSR);
    if( (iir_stat & UART_LSR_DR) != 0 ) return 2; //info_port2
    else return 0;                //lozhnoe preryvanie
#endif
}

void rs485_allow_interrupt_card(RS485INFO *pinfo) {
//    
    outb(UART_IER_RDI, pinfo->base +UART_IER);
}

void rs485_stop_interrupt_card(RS485INFO *pinfo) {
//    
    outb(0, pinfo->base +UART_IER);
}

inline void rs485_read_enable(RS485INFO *pinfo) {
    outb(RXENABLE,pinfo->base +RX_TX_REG ); //  
}

inline void rs485_transmit_enable(RS485INFO *pinfo) {
    outb(TXENABLE,pinfo->base +RX_TX_REG ); //  
}

int rs485_test_interface(RS485INFO *pinfo) {
    rs485_read_enable(pinfo);
//    udelay(4000); // 4 msec

#ifdef __COM__
    if( (inb(pinfo->base +UART_MSR) & UART_MSR_CTS) == 0 ) 
                return -ERSNOINTERFACE; // no COM-rs485 converter
#else
    if( RXENABLE != inb(pinfo->base +RX_TX_REG) )
                return -ERSNOINTERFACE; // no moxa102
#endif
    return SUCCESS;
}

int rs485_test_power(RS485INFO *pinfo) {
#ifdef __COM__
// Echo test:
    rs485_transmit_enable(pinfo);
    outb_p(0x1f, pinfo->base);  // test byte to test address
    rs485_set_starttime();
    while( 1 ) {
        if( ( (inb_p(pinfo->base +UART_LSR)) & UART_LSR_DR ) != 0 ) break;
        if( rs485_timeout(TBYTE_TIMEOUT) ) {
            rs485_read_enable(pinfo);
	    PDEBUG( "test_device: timeout: no echo!\n" );
	    return -ERSSHUTDOWN;   // no echo => no power
        }
        udelay(TBYTE_WAIT);
    }
    if(inb_p(pinfo->base) != 0x1f) PDEBUG("BAD ECHO!!!\n");
    rs485_read_enable(pinfo);
#endif
    return SUCCESS;
}

int rs485_test_device(RS485INFO *pinfo) {
    int ret = SUCCESS;
    if( (ret=rs485_test_interface(pinfo)) != SUCCESS ) return ret;;
    return rs485_test_power(pinfo);
}

void rs485_clean_reg(RS485INFO *pinfo) {
    unsigned short i;
    for(i=0; i<8; i++) outb_p(0x0, pinfo->base+i);
}

#ifdef __COM__
    #define Fosc        1843
    #define Minbaudrate    19
    #define Maxbaudrate    115
#else
    #define Fosc        14746
    #define Minbaudrate    115
    #define Maxbaudrate    921
#endif

int ubrr_to_rate(int baudrate_code) {
    if(baudrate_code == 0) return 0;
    return(Fosc/16/baudrate_code);
}

int rate_to_ubrr(int baudrate) {
    int baudrate_code;
    baudrate = baudrate < Minbaudrate ? Minbaudrate : baudrate;
    baudrate = baudrate > Maxbaudrate ? Maxbaudrate : baudrate;
//    baudrate_code = iround(Fosc/16.0/baudrate);
    baudrate_code = Fosc/16/baudrate;
    return baudrate_code;
}

int rs485_set_baudrate(RS485INFO *pinfo) {
//  
    int baudrate_code;
    baudrate_code = rate_to_ubrr(pinfo->baudrate);
    pinfo->baudrate = ubrr_to_rate(baudrate_code);
    PDEBUG("set baudrate=%d\n",pinfo->baudrate);
    outb_p(0x80, pinfo->base +UART_LCR);
    outw(baudrate_code, pinfo->base +UART_DLL); 
    outb_p(OTHER, pinfo->base +UART_LCR);    // .  
    return SUCCESS;
}

int rs485_reset(RS485INFO *pinfo) {    //  -


    rs485_clean_reg(pinfo);
    rs485_read_enable(pinfo);
    udelay(4000); // 4 msec
    rs485_allow_interrupt_card(pinfo);

    rs485_set_baudrate(pinfo);

//    outb_p(OTHER, pinfo->base +UART_LCR);    // .  

    outb_p(UART_FCR_ENABLE_FIFO
                    | UART_FCR_CLEAR_RCVR
                    | UART_FCR_CLEAR_XMIT
                    | UART_FCR_TRIGGER_8    //   fifo,
                                            //     - 
                                            //    1- 
                    ,pinfo->base +UART_FCR);  // fifo

//   
 //    inb_p( pinfo->base +UART_LSR ); // eto schityvanie nichego krome PE-bita ne meniayet
    inb_p( pinfo->base +UART_RX );

//    rs485_fifo_flush(pinfo);
    return rs485_test_device(pinfo);
}

// -----------------------------------------------------------------------

int start_transmitter(RS485INFO *info) {
/*   
 *   < 0
 *   > CACHE_MAXSIZE0
 *      < CACHE_MAXSIZE
*/
    unsigned  short iir=0;
    int ret=0;

    int rbyte_tmout=0;

//   mutex-  -  ( )???
//     -   =>    ???
//     ( ).   lpt ?
    rs485_stop_interrupt_card(info);
    rs485_stop_interrupts();    // chtoby ne bylo pause ot handlerov

    outb_p(FIRST, info->base +UART_LCR);    //  "   "
    rs485_transmit_enable(info);
    outb_p(info->outbuffer.cache[0], info->base); //   

    info->usHopOut = 0;    //  hop
    info->usHopInp = 0;    //  hop
    info->signal = 0;

    //   
    info->usCountByteOut = 0;    //    0
    info->usCountByteInp = 0;    //    0
    info->lSection = SECT_ADDRESSBYTE;    //   

    while( 1 ) {

        //  
        rs485_set_starttime();
        rbyte_tmout = 0;
        while( 1 ) {
            if( ( (iir=inb_p(info->base +UART_LSR)) & UART_LSR_DR ) != 0 ) break;
            if( rs485_timeout(TBYTE_TIMEOUT) ) {
                rbyte_tmout = 1;
                break;
            }
            udelay(TBYTE_WAIT);
        }


//       

        if( rbyte_tmout ) {

            switch(info->lSection) {

// no echo:
                case SECT_ADDRESSBYTE :
                case SECT_SIGNAL      : 
                    outb_p(OTHER, info->base +UART_LCR); // . 
                case SECT_OUTBYTES    :
                    rs485_read_enable(info);
                    ret = -ERSSHUTDOWN;   // no echo => no power
                    PDEBUG( "transmit: timeout: no echo!\n" );
                    goto QUIT_TRANSMITTER;    
// no reply:
                case SECT_REPLYBYTE    :
                    ret = SignalDSC;
                    PDEBUG( "transmit: timeout: no reply!\n" );
                    goto QUIT_TRANSMITTER;

                case SECT_BADATA        :    //   bad-bytes,
                                    //  
                    
                    info->lSection = SECT_DATABYTES;
                    ret = SignalEER;
                    PDEBUG("transmit: end of BAD DATA packet (timeout)\n");

// data loss or FE:
                case SECT_DATABYTES    :    //     
                    (info->usHopInp) ++;    //   hop
                    if(info->usHopInp < MAX_HOP) {
                        PDEBUG( "transmit: broken packet!\n" );
                        info->signal = SignalNAK;
                    } else { 
                        //  HOP   ACK   
//   ACK -???
                        PDEBUG( "transmit: broken packet, MAX_HOP reached!\n" );
                        ret = SignalEER;
                        info->signal = SignalACK;
                    }
                    goto SEND_SIGNAL;

                default    :    
                    goto QUIT_TRANSMITTER;
            }
        }    // end of timeout-condition

        info->tmp_byte = inb_p(info->base); //  
//        RS_PRINT( ">%X\n", info->tmp_byte );

        if( (iir & UART_LSR_FE) != 0 ) { //  .  , 
            PDEBUG( "transmit: frame error!\n" );

            switch(info->lSection) {

                case SECT_ADDRESSBYTE :
                case SECT_SIGNAL      : 
                    outb_p(OTHER, info->base +UART_LCR);    // . 
                case SECT_OUTBYTES    :
                    rs485_read_enable(info);
                    ret = SignalEER;
                    goto QUIT_TRANSMITTER;
                    break;

                case SECT_REPLYBYTE   :
                case SECT_DATABYTES   :
                    info->lSection = SECT_BADATA; // !!!       -
                    break;
                case SECT_BADATA      :
                    break;
                default: break;
            }
        }

        if(        //   -   
            (
                ( (info->lSection == SECT_ADDRESSBYTE)
                || (info->lSection == SECT_OUTBYTES) )
                && (info->tmp_byte != *(info->outbuffer.cache +info->usCountByteOut))
            )
            || 
            ( (info->lSection == SECT_SIGNAL)
                && (info->tmp_byte != info->signal)
            )
        ) {

            //        -
            //      -ECONNREFUSED
            // -   ("COL"  TURBINA)

            if( (info->lSection == SECT_ADDRESSBYTE)
                || (info->lSection == SECT_SIGNAL) ) {
                outb_p(OTHER, info->base +UART_LCR);    // . 
                PDEBUG( "transmit: bad Address or Signal echo!\n" );
            } else {
                PDEBUG( "transmit: bad echo (non Address or Signal)!\n" );
            }
            rs485_read_enable(info);
            ret = SignalCOL;
            goto QUIT_TRANSMITTER;
        }

        switch( info->lSection ) { //    
        
            case SECT_ADDRESSBYTE:

                //   
                //        OUTBYTES
                outb_p(OTHER, info->base +UART_LCR);
                info->lSection = SECT_OUTBYTES;

            case SECT_OUTBYTES:

                if( info->usCountByteOut == info->outbuffer.size ) {
                    //   
                    //    REPLYBYTE
                    //   

                    rs485_read_enable(info);
                    info->lSection = SECT_REPLYBYTE;

                } else {

                    //     
                    (info->usCountByteOut) ++;
                    outb_p(*(info->outbuffer.cache +info->usCountByteOut),
                             info->base);
                }
                break;

            case SECT_REPLYBYTE: 

                //  

                if( iir  & UART_LSR_PE ) {    //  

// !!! Iz-za pomekh na linii vmesto ACY inogda prikhisit ACK, t.k. oni
// !!! otlichayutsia tol'ko odnim bitom.
// !!! Tut ACK ne mozhet byt', znachit, eto iskazhennyi ACY :-)
//                    if( info->tmp_byte == SignalACK ) info->tmp_byte = SignalACY;

                    if( info->tmp_byte != (unsigned char) SignalNAK ) { 

                        //     nak  
                        //     
                        ret = info->tmp_byte;
                        goto QUIT_TRANSMITTER;

                    } else {    //   nak

                        PDEBUG( "transmit: NAK from module!\n" );
                        info->usHopOut ++;
                        if( info->usHopOut < MAX_HOP ) {

                            //       
                            //   

                            outb_p(FIRST, info->base +UART_LCR); //  1- 
                            rs485_transmit_enable(info);
                            info->usCountByteOut = 0;
                            info->lSection = SECT_ADDRESSBYTE;

                            //  
                            outb_p(info->outbuffer.cache[0], info->base);
                            break;

                        } else {

                            //     
                            ret = SignalNAK;
                            goto QUIT_TRANSMITTER;
                        }
                    }

                } else { //    

                    //  1   
                    if( info->tmp_byte < (unsigned char)0x20 ) {

                        info->passive_inbuf.size = info->tmp_byte +2;
                        info->passive_inbuf.cache[info->usCountByteInp]=info->tmp_byte;
                        info->lSection = SECT_DATABYTES;
                        (info->usCountByteInp) ++;

                    } else {
                        PDEBUG( "transmit: bad length-byte, skipped!\n" );
                        //     ,
                        //    (NAK)
                        //         (FE, )
                        info->lSection = SECT_BADATA;
                    }
                }
                break;

            case SECT_BADATA:
                break;

            case SECT_DATABYTES: 
            
                //  

                info->passive_inbuf.cache[info->usCountByteInp] =
                            info->tmp_byte; //    
                (info->usCountByteInp) ++;    

                if( info->usCountByteInp == info->passive_inbuf.size ) {

                    //    -   
                    if( crc(info->passive_inbuf.cache, 
                            info->passive_inbuf.size) != 0 ) {

                        //   
                        PDEBUG( "transmit: bad CRC!\n" );
                        (info->usHopInp) ++;    //   hop
                        if(info->usHopInp == MAX_HOP) { 

                            //  HOP  ACK   

                            info->signal = SignalACK;
                            ret = SignalEER;

                        } else info->signal = SignalNAK; //   NAK
                    } else { 
                        //   
                        //  ACK       

                        info->signal = SignalACK;
                        ret = info->passive_inbuf.size-1;    // size of data without CRC
                    }
                    goto SEND_SIGNAL;
                }
                break;

            case  SECT_SIGNAL:

                //    
                //    

                outb_p(OTHER, info->base +UART_LCR);
                rs485_read_enable(info);

                if(info->signal == SignalNAK) {
                    //     NAK,     
                    info->signal = 0;
                    info->lSection = SECT_REPLYBYTE;
                    info->usCountByteInp = 0;
                } else {
                    // ACK
                    goto QUIT_TRANSMITTER;
                }
                break;

            default: break;

        }    //   

SEND_SIGNAL:

//        if( info->signal && (info->lSection == SECT_DATABYTES) ) {
        if(info->signal) {

            //    

            outb_p(FIRST, info->base +UART_LCR); //   
            rs485_transmit_enable(info);
            info->lSection = SECT_SIGNAL;
            outb_p(info->signal, info->base); 
        }

    }    //    

QUIT_TRANSMITTER:

    udelay(TBYTE_WAIT);

    rs485_allow_interrupts();
    rs485_allow_interrupt_card(info);
    return ret;
}

void receiv_data(RS485INFO * info) {

/*
 * -----------------------------------------------------------
 *          
 * ------------------------------------------------------------
*/

    int rbyte_tmout = 0;
    unsigned short iir = 0;
#ifndef __COM__
    int first = 1;
#endif

    static int SharedMaskInterrupt;
    SharedMaskInterrupt = UART_LSR_DR;


    rs485_stop_interrupt_card(info);
#ifdef __DEBUG_TIMING__
    rs485_set_starttime();
#endif


#ifdef __DEBUG_SIGNAL__
    outb_p(TXENABLE, info->base +RX_TX_REG + 8);
#endif

    info->signal = 0;
    info->lSection = SECT_ADDRESSBYTE;
    info->usCountByteInp = 0;
    info->active_inbuf.size = 0;
    info->usHopInp = 0;

    while( 1 ) {

//    
        rs485_set_starttime();
        rbyte_tmout = 0;
#ifndef __COM__
        if(first != 1) { // gotovnost' uzhe schitana pri vhode v get_port
#endif        
        while( 1 ) {
            iir = inb(info->base +UART_LSR);
            if( (iir & SharedMaskInterrupt) != 0 ) break;

            if( rs485_timeout(RBYTE_TIMEOUT) ) {
                rbyte_tmout = 1;
                break;
            }
            udelay(RBYTE_WAIT);
        }
#ifndef __COM__
        } else {
            first = 0;
            iir = iir_stat;
        }
#endif

//   - - 
        if( rbyte_tmout ) {
            switch(info->lSection) {
                case SECT_SIGNAL :
                    PDEBUG( "receiver: no signal-echo, time-out!\n" );
                    outb_p(OTHER, info->base +UART_LCR);
                    rs485_read_enable(info);
                    goto QUIT_HANDLER;

                case SECT_ADDRESSBYTE:
                case SECT_LENGTHBYTE:
                    goto QUIT_HANDLER;

                case SECT_BADATA :
                    PDEBUG( "receiver: end of BAD DATA packet, time-out!\n" );
                    info->lSection = SECT_DATABYTES;
// data loss or FE:
                case SECT_DATABYTES    :    //     
                    (info->usHopInp) ++;    //   hop
                    if(info->usHopInp < MAX_HOP) {
                        PDEBUG( "receiver: broken packet!\n" );
                        info->signal = SignalNAK;
                    } else { 
                        PDEBUG( "receiver: broken packet, MAX_HOP reached!\n" );
                        info->signal = SignalACK;
                    }
                    goto SEND_SIGNAL;


               default           :
                    PDEBUG( "receiver: time-out! section=%ld\n", info->lSection );
                    goto QUIT_HANDLER;
            }
        }

//  
        info->tmp_byte = inb_p(info->base);
//            RS_PRINT( "%02X ",info->tmp_byte );
            
//   ? 
        if( (iir & UART_LSR_FE) != 0 ) {
            PDEBUG( "receiver: frame error!\n" );
//      -
            if( info->lSection == SECT_SIGNAL ) {
                outb_p(OTHER, info->base +UART_LCR);
                rs485_read_enable(info);
                goto QUIT_HANDLER;
            } else info->lSection = SECT_BADATA;
        }

//   FIFO?
        if( (iir & UART_LSR_OE) != 0 ) {

            info->active_inbuf.signature |= S_FIFO_OVERFLOW;
            PDEBUG( "receiver: hard-FIFO overflow!\n" );

            //  FIFO
            outb_p( UART_FCR_ENABLE_FIFO | UART_FCR_CLEAR_RCVR
                | UART_FCR_CLEAR_XMIT
                | UART_FCR_TRIGGER_8    //   fifo,
                                        //     - 
                                        //    1- 
                ,info->base +UART_FCR );

            info->lSection = SECT_BADATA;
        }

        switch( info->lSection ) {

        case SECT_BADATA :
            break;

        case SECT_SIGNAL :
            outb_p(OTHER, info->base +UART_LCR);    //    "  "
            rs485_read_enable(info);
            if( info->tmp_byte != info->signal ) {
                PDEBUG( "receiver: signal: bad echo!\n" );
//                  - ...
                goto QUIT_HANDLER;
            }
            if(info->signal == SignalNAK) {
                    //     NAK,     
                info->signal = 0;
                info->usCountByteInp = 0;
                info->lSection = SECT_ADDRESSBYTE;
            } else { // ACK
                goto QUIT_HANDLER;
            }
            break;
            //    ,     ...

        case SECT_ADDRESSBYTE:

            //        
            //   ,    (  :-))
            //    - ,     
            if((unsigned char)(info->tmp_byte) >= (unsigned char) CACHE_MAXSIZE ) {
                PDEBUG( "receiver: signal %x in address-section!\n",info->tmp_byte );
                info->lSection = SECT_BADATA;
                break;
            }
            //   ,     
            info->active_inbuf.cache[info->usCountByteInp] = info->tmp_byte;
            info->usCountByteInp ++;
            info->active_inbuf.adr = info->tmp_byte;
            info->lSection = SECT_LENGTHBYTE;
            break;

        case SECT_LENGTHBYTE:
            if(info->tmp_byte >= 0x20) {
                PDEBUG( "receiver: length error!\n" );
                info->lSection = SECT_BADATA;
                break;
            }
            info->active_inbuf.cache[info->usCountByteInp] = info->tmp_byte;
            info->usCountByteInp ++;
            info->active_inbuf.size = info->tmp_byte+3;
            info->lSection = SECT_DATABYTES;
            break;

        case SECT_DATABYTES:            
            //     or  - 
            info->active_inbuf.cache[info->usCountByteInp] = info->tmp_byte;
            info->usCountByteInp ++;
            if( info->usCountByteInp < info->active_inbuf.size ) break;
            
            if( (crc(info->active_inbuf.cache, info->active_inbuf.size)) != 0 ) {
                PDEBUG( "receiver: bad CRC!\n" );
                if( (info->usHopInp) < MAX_HOP ) {
                    //    hop- -  NAK    
                    (info->usHopInp) ++;
                    info->signal = SignalNAK;
                } else {
                    //     hop-,     
//                    info->active_inbuf.signature |= S_DATA_NAK;
                    info->signal = SignalACK;
                }
                goto SEND_SIGNAL;
            }
#ifdef __DEBUG_TIMING__
            if(rs485_timeout(100L)) RS_PRINT("Long packet > 100 usec!\n");
#endif
            
            rs485_put_data(info);
            //    ACK
            info->signal = SignalACK;
            goto SEND_SIGNAL;
            break;

        }  // end of switch
SEND_SIGNAL:
        if( info->signal ) {
            //  
            outb_p(FIRST, info->base +UART_LCR);    //   "  "
            rs485_transmit_enable(info);
            outb_p(info->signal, info->base);
            info->lSection = SECT_SIGNAL;
        }
    } //   

QUIT_HANDLER:

#ifdef __DEBUG_SIGNAL__
    outb_p(RXENABLE, info->base +RX_TX_REG + 8);
#endif
    rs485_allow_interrupt_card(info);

    return;
}
