#include <rs485.h>
#include <mcua.h>
#include <linux/interrupt.h>
#include <asm/irq.h>
#ifdef __CP132__
	#include <moxa132.h>
#endif
#ifdef __CP102__
	#include <moxa102.h>
        #include "moxa102_com.h"
#endif
#ifdef __COM__
	#include <com.h>
        #include "moxa102_com.h"
#endif
#ifdef __LPT__
	#include <lpt.h>
#endif
#include "rs485_irq.h"

extern RS485INFO *info_port1, *info_port2;

//    
#ifndef KERN_2_6
static void rs485_interrupt(int irq, void *dev_id, struct pt_regs *regs) {
#else
static irqreturn_t rs485_interrupt(int irq, void *dev_id, struct pt_regs *regs) {
#endif
	RS485INFO *info_port = info_port1;
	int res;

	res = rs485_get_port(irq);

	if(res == 1) info_port = info_port1;
	else if(res == 2) info_port = info_port2;
	else {
		PDEBUG("rs485_tools: empty interrupt!\n");
#ifndef KERN_2_6
		return;
#else
		return IRQ_NONE;
#endif
	}
	receiv_data(info_port);
#ifndef KERN_2_6
	return;
#else
	return IRQ_HANDLED;
#endif
}

static int dev_id = 1;
int rs485_request_irq(unsigned int irq) {
	int err=0;
#ifndef KERN_2_6
	unsigned long flags;
	save_flags(flags); cli();
#endif
// SA_SHIRQ - PCI trebuet shared irq (2.6 -tochno, 2.4 - ne znayu)
	if( (err=request_irq(irq, rs485_interrupt,
				SA_INTERRUPT 
#ifdef __PCI__
				| SA_SHIRQ
#endif
				, ADAPTER_ID,
				&dev_id)) != 0 ) {
		RS_PRINT(KERN_CRIT "rs485: request irq %d failed\n",irq);
	}
#ifndef KERN_2_6
	sti(); restore_flags(flags);
#endif

    return err;

}

void rs485_clean_irq(unsigned int irq) {
#ifndef KERN_2_6	
	unsigned long flags;
	save_flags(flags); cli();
	free_irq(irq, &dev_id);
	sti(); restore_flags(flags);
#else
	free_irq(irq, &dev_id);
#endif
}


static unsigned long flags;
static int is_flags_saved = 0;

#ifndef KERN_2_6
void rs485_stop_interrupts(void) {
// disable all interrupts
	save_flags(flags);
	is_flags_saved = 1;
	cli();
}

void rs485_allow_interrupts(void) {
// enable interrupts
	sti();
	if(is_flags_saved) restore_flags(flags);
}
#else
void rs485_stop_interrupts(void) {
	is_flags_saved = 1;
	local_irq_save(flags);
}

void rs485_allow_interrupts(void) {
	if(is_flags_saved) local_irq_restore(flags);
}
#endif

